package be.iminds.ilabt.jfed.rspec.model;

import be.iminds.ilabt.jfed.rspec.parser.EventListExtraXml;
import be.iminds.ilabt.jfed.rspec.parser.RspecParseException;
import be.iminds.ilabt.jfed.util.common.GeniUrn;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.xml.stream.XMLEventWriter;
import javax.xml.stream.XMLStreamException;
import java.util.Date;

/**
 * RspecNodeFactory
 */
public interface RspecFactory {
    @Nonnull ModelRspecType getModelRspecType();

    @Nonnull RspecNode createNode(@Nonnull ModelRspec modelRspec);
    @Nonnull RspecNode createNodeWithClientId(@Nonnull ModelRspec modelRspec, @Nullable String clientId);
    @Nonnull RspecNode copyNode(@Nonnull ModelRspec modelRspec, @Nonnull RspecNode rspecNode, @Nonnull RspecNode.InterfaceCopyMethod interfaceCopyMethod);
    @Nonnull RspecNode copyNodeIncludingInterfaces(@Nonnull ModelRspec newModelRspec, @Nonnull RspecNode node);

    @Nonnull GeantTestbedType copyGeantTestbedType(@Nonnull ModelRspec modelRSpec, @Nonnull GeantTestbedType gtt);

    /**
     * Based on extraXml and/or other properties, convert this RSpecNode into a more specialized RspecNode class (which also implements the RspecNode interface).
     *
     * @param origNode
     * @return the same node, or a new node of a more specialised class that contains the same info.
     */
    @Nonnull RspecNode convertToSpecialNode(@Nonnull RspecNode origNode) throws RspecParseException;

    @Nonnull RspecLink createLink(@Nonnull ModelRspec rspec, @Nullable String addDefaultLinkType);
    @Nonnull RspecLink createLinkWithClientId(@Nonnull ModelRspec rspec, @Nonnull String id, @Nullable String addDefaultLinkType);
    @Nonnull RspecLink copyLink(@Nonnull ModelRspec newModelRspec, @Nonnull RspecLink rspecLink, @Nonnull RspecNode.InterfaceCopyMethod interfaceCopyMethod);

    @Nonnull RspecInterface createInterface(@Nonnull RspecNode node, @Nullable String id);
    @Nonnull RspecInterface createInterface(@Nonnull RspecLink link, @Nullable String id);
    @Nonnull RspecInterface createInterface(@Nonnull RspecNode node, @Nonnull RspecLink link, @Nullable String id);
    /* choose an interface name automatically */
    @Nonnull RspecInterface createInterface(@Nonnull RspecNode newRspecNode, @Nonnull RspecLink link);

    @Nonnull LinkSetting createLinkSetting(@Nonnull RspecInterface fromIface, @Nonnull RspecInterface toIface);

    @Nonnull ExecuteService createExecuteService(@Nullable String shell, @Nullable String command);
    @Nonnull InstallService createInstallService(@Nullable String installPath, @Nullable String url);
    @Nonnull AnsibleService createAnsibleService();

    @Nonnull ModelRspec createModelRspec(@Nullable String rspecType);
    @Nonnull ModelRspec copyModelRspec(@Nonnull ModelRspec orig, @Nullable String newType);


    /**
     * called when the parser encounters extra XML in a node.
      * @return true if the extr aXML was processed and does not need to be stored in extraXml. False otherwise.
     */
    boolean processExtraRspecXml(@Nonnull ModelRspec model, @Nonnull EventListExtraXml extraXml) throws RspecParseException;
    /**
     * called when the parser encounters extra XML in a node.
      * @return true if the extr aXML was processed and does not need to be stored in extraXml. False otherwise.
     */
    boolean processExtraNodeXml(@Nonnull RspecNode node, @Nonnull EventListExtraXml extraXml) throws RspecParseException;
    /**
     * called when the parser encounters extra XML in a node.
      * @return true if the extr aXML was processed and does not need to be stored in extraXml. False otherwise.
     */
    boolean processExtraLinkXml(@Nonnull RspecLink link, @Nonnull EventListExtraXml extraXml) throws RspecParseException;

    void writeExtraRspecXml(@Nonnull ModelRspec modelRspec, @Nonnull XMLEventWriter eventWriter) throws XMLStreamException;
    void writeExtraNodeXml(@Nonnull RspecNode node, @Nonnull XMLEventWriter eventWriter) throws XMLStreamException;
    void writeExtraLinkXml(@Nonnull RspecLink link, @Nonnull XMLEventWriter eventWriter);


    //must have equals and hashcode  (they are very simple as all factorries of the same class will be equal.
    // This is needed because RspecSource uses RspecFactory as HashMap key.
    boolean equals(Object o);
    int hashCode();

    /** merge extra XML into the merged model. Used for merging BarrierSegments */
    void mergeExtra(@Nonnull ModelRspec mergedModel, @Nonnull ModelRspec partialRspec);

    /** called when parsing is finished, in case the factory wants to do some last change in the model.
     * Currently used for adding commands to barriers. */
    void finishParsing(@Nonnull ModelRspec modelRspec);

    @Nonnull OpenflowDataPath createOpenflowDataPath(@Nonnull ModelRspec rspec, @Nonnull EventListExtraXml allXml, @Nullable String componentManagerId, @Nullable String componentId, @Nullable String dpidAttValue);

    @Nonnull Lease createLease(@Nonnull ModelRspec rspec, @Nullable String leaseId, @Nullable String clientId, @Nullable Date validFromDate, @Nullable Date validUntilDate, @Nullable String validFromString, @Nullable String validUntilString);
    @Nonnull Lease createLease(@Nonnull ModelRspec rspec, @Nullable String leaseId, @Nullable String clientId, @Nullable Date validFromDate, @Nullable Date validUntilDate);

    @Nonnull Channel createChannel(@Nonnull ModelRspec modelRspec);

    @Nonnull GeantTestbedType createGeantTestbedType(@Nullable String name);

    @Nonnull AddressPool createAddressPool(@Nonnull ModelRspec modelRspec);
    @Nonnull AddressPool copyAddressPool(@Nonnull ModelRspec modelRspec, @Nonnull AddressPool addressPool);
    @Nonnull IPv4 createIPv4(@Nonnull ModelRspec modelRspec);
    @Nonnull IPv4 createIPv4(@Nonnull ModelRspec modelRspec, @Nullable String address, @Nullable String netmask);

    @Nonnull ExecuteAnsiblePlaybook createExecuteAnsibleCookbook(@Nonnull ModelRspec rspec, @Nullable String source, @Nullable String outputFilename);

    @Nonnull DistributeSshKeypair createDistributeSshKeypair(@Nonnull ModelRspec rspec, @Nullable String location, @Nullable GeniUrn userUrn);

    @Nonnull GeantTestbedType.VpnService createGtsVpnService(String endpoint, String username, String password);
    @Nonnull WebVncService createWebVncService(String providerId, String vncUrl, String username, String password);
}
